"""
Utility functions. Mostly math stuff.
"""

import numpy as np
import scipy as sc


# Returns the top eigenvectors (according to increasing eigenvalues) of a symmetric matrix.
def eigenvectors(matrix, num=1):
    return sc.linalg.eigh(matrix, eigvals=(0, num - 1))[1]


# Returns the smallest eigenvalues of a symmetric matrix.
def eigenvalues(matrix, num=1):
    return sc.linalg.eigh(matrix, eigvals=(0, num - 1))[0]


# Returns the normalized Laplacian for a given similarity matrix.
def normalized_laplacian(similarity_matrix):
    A = (similarity_matrix + similarity_matrix.T)/2
    R = np.sum(A, axis=1)
    R_minus_half = 1/np.sqrt(R)
    D_minus_half = np.diag(R_minus_half)
    I = np.eye(A.shape[0])
    return I - D_minus_half * A * D_minus_half


# Returns the unnormalized Laplacian for a given similarity matrix.
def unnormalized_laplacian(similarity_matrix):
    A = (similarity_matrix + similarity_matrix.T)/2
    R = np.sum(A, axis=1)
    D = np.diag(R)
    return D - A


# Returns the pairwise distance between the row-wise vectors (representing samples).
def distance_matrix(matrix, metric):
    return sc.spatial.distance.squareform(sc.spatial.distance.pdist(matrix, metric=metric))


# Solves the sparse simplex reconstruction problem for a vector, v.
# Follows notation given in 'A New Simplex Sparse Learning Model to Measure Data Similarity for Clustering', IJCAI 2015.
def best_simplex_projection(vector, num_iterations=100):

    # Compute quantities.
    v = np.array(vector)
    n = v.size
    u = v - np.mean(v) + np.ones(n)/n

    if np.min(u) < 0:
        
        func = 1
        multiplier = 0
        iteration = 0

        while np.abs(func) > 1e-10 and iteration < num_iterations:
            alpha_uncut = u - multiplier
            indices = (alpha_uncut > 0)
            
            # Compute function value and derivative.
            func_deriv = -np.sum(indices)
            func = np.sum(alpha_uncut[indices]) - 1
            
            # Newton's rule update.
            multiplier = multiplier - func/func_deriv    

        alpha = np.maximum(alpha_uncut, 0)

    else:
        alpha = u
    
    return alpha



from itertools import product
from collections import defaultdict
import numpy as np

class PathSim:
    def __init__(self, type_lists, incidence_matrices):
        self.node_types = type_lists
        self.node_type_indices = {node: index for type_list in type_lists.values() for index, node in enumerate(type_list)}
        self.incidence_matrices = incidence_matrices
        self.fill_incidence_matrices()
        self.similarity_matrices = {}
    
    # Transpose for the reversed incidence type.
    def fill_incidence_matrices(self):
        initial_matrices = self.incidence_matrices.copy()
        for incidence_type in initial_matrices.keys():
            self.incidence_matrices[incidence_type[::-1]] = self.incidence_matrices[incidence_type].T

    # Find similarities between node1 and node2.
    def pathsim(self, node1, node2, metapath):
                    
        if metapath not in self.similarity_matrices:
            self.compute_similarity_matrix(metapath)
        
        try:
            return self.similarity_matrices[metapath][self.node_type_indices[node1]][self.node_type_indices[node2]]
        except ValueError:
            raise ValueError('Invalid node types for given metapath.')
        
    # Computes the similarity matrix by iterating over all pairs of nodes.
    def compute_similarity_matrix(self, metapath):

        # Compute metapath matrix.
        metapath_matrix = self.compute_metapath_matrix(metapath)

        # Fill in similarity matrix entries now.
        num_nodes = len(self.node_types[metapath[0]])
        similarity_matrix = np.eye(num_nodes, num_nodes)

        for (index1, index2) in product(range(num_nodes), range(num_nodes)):
            if index1 != index2:
                num_paths_12 = metapath_matrix[index1][index2]
                num_paths_11 = metapath_matrix[index1][index1]
                num_paths_22 = metapath_matrix[index2][index2]

                if (num_paths_11 + num_paths_22) > 0:
                    similarity_matrix[index1][index2] = (2 * num_paths_12)/(num_paths_11 + num_paths_22)
                else:
                    similarity_matrix[index1][index2] = 0
        
        self.similarity_matrices[metapath] = similarity_matrix
        return similarity_matrix

    # Computes the number of paths via this metapath.
    def compute_metapath_matrix(self, metapath):

        # We only support symmetric metapaths, for now.
        if not PathSim.symmetric(metapath):
            raise ValueError('Only symmetric metapaths supported.')
            
        curr_matrix = np.eye(len(self.node_types[metapath[0]]))
        metapath_length = len(metapath)

        for node_type, next_node_type in zip(metapath[:metapath_length//2], metapath[1:metapath_length//2 + 1]):
            
            incidence_type = node_type + next_node_type

            if incidence_type not in self.incidence_matrices.keys():
                raise ValueError('Invalid incidence %s.' % incidence_type)

            curr_matrix = np.matmul(curr_matrix, self.incidence_matrices[incidence_type])

        return np.matmul(curr_matrix, curr_matrix.T)

    # Check if metapath is symmetric. 
    @staticmethod
    def symmetric(metapath):
        metapath_length = len(metapath)
        for index, node_type in enumerate(metapath[:metapath_length//2]):
        
            other_index = metapath_length - index - 1
            other_node_type = metapath[other_index]

            if other_node_type != node_type:
                print(other_node_type, node_type)
                return False

        return True
        
    

        






"""
A Python Implementation of 'Spectral Clustering in Heterogeneous Information Networks' by Xiang Li, Ben Kao, Zhaochun Ren, Dawei Yin, at AAAI 2019.
"""

# External dependencies.
from sklearn.cluster import KMeans
import numpy as np
from scipy.optimize import minimize

# Internal dependencies.
try:
    from .utils import eigenvalues, eigenvectors, normalized_laplacian, distance_matrix, best_simplex_projection
except (ImportError, ValueError, SystemError):
    from utils import eigenvalues, eigenvectors, normalized_laplacian, distance_matrix, best_simplex_projection

class SClump:
    def __init__(self, similarity_matrices, num_clusters, random_seed=0):
        
        self.num_clusters = num_clusters
        self.random_seed = random_seed
        self.num_nodes = None
        self.similarity_matrices = []
        self.metapath_index = {}

        # Loss function coefficients. We'll fill this in later.
        self.alpha = None
        self.beta = None
        self.gamma = None

        for index, (metapath, matrix) in enumerate(similarity_matrices.items()):
            if self.num_nodes is None:
                self.num_nodes = matrix.shape[0]
            
            if matrix.shape != (self.num_nodes, self.num_nodes):
                raise ValueError('Invalid shape of similarity matrix.')
            
            # Normalize so that row sums are 1.
            row_normalized_matrix = matrix/matrix.sum(axis=1, keepdims=True)
            
            self.similarity_matrices.append(row_normalized_matrix)
            self.metapath_index[metapath] = index

        self.similarity_matrices = np.array(self.similarity_matrices)
        self.num_metapaths = len(similarity_matrices)
        

    def run(self, verbose=False, cluster_using='similarity'):
        """
        Returns a tuple of:
        * labels: The predicted cluster labels for each node.
        * similarity: The learnt similarity matrix, from the optimization procedure.
        * metapath_weights: A dictionary of learnt weights for each metapath's associated similarity matrix.
        """
        if cluster_using not in ['similarity', 'laplacian']:
            raise ValueError('Invalid option for parameter \'cluster_using\'.')

        similarity_matrix, metapath_weights = self.optimize(verbose=verbose)

        if cluster_using == 'similarity':
            labels = self.cluster(similarity_matrix)

        elif cluster_using == 'laplacian':
            laplacian = normalized_laplacian(similarity_matrix)
            labels = self.cluster(eigenvectors(laplacian, num=self.num_clusters))

        metapath_weights_dict = {metapath: metapath_weights[index] for metapath, index in self.metapath_index.items()}
        
        return labels, similarity_matrix, metapath_weights_dict
    
    
    def cluster(self, feature_matrix):
        """
        Cluster row-wise entries in the feature matrix. Currently, k-means clustering is used.
        TODO: Add spectral rotation as an alternative to k-means.
        """
        return KMeans(self.num_clusters, n_init=10, random_state=self.random_seed).fit_predict(feature_matrix)


    def optimize(self, num_iterations=20, alpha=0.5, beta=10, gamma=0.01, verbose=False):
        """
        Learn weights to optimize the similarity matrix.
       
        Coefficients in the loss function:
        * alpha: for the Frobenius norm of S, 
        * beta: for the L2-norm of lambda, 
        * gamma: for the trace of LS.
        """

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

        # Weights over similarity matrices.
        lambdas = np.ones(self.num_metapaths)/self.num_metapaths
        W = np.tensordot(lambdas, self.similarity_matrices, axes=[[0], [0]])
        
        # Initialize.
        S = W

        # Iterate.
        for iteration in range(num_iterations):
            
            if verbose:
                loss = np.trace(np.matmul((S - W).T, (S - W))) 
                loss += self.alpha * np.trace(np.matmul(S.T, S))
                loss += self.beta * np.dot(lambdas, lambdas)
                loss += self.gamma * np.sum(eigenvalues(normalized_laplacian(S), num=self.num_clusters))

                print('Iteration %d: Loss = %0.3f' % (iteration, loss))

            # Update F.
            F = self.optimize_F(S)

            # Update S.
            S = self.optimize_S(W, F)
            
            # Update lambdas.
            lambdas = self.optimize_lambdas(S, lambdas)

            # Recompute W.
            W = np.tensordot(lambdas, self.similarity_matrices, axes=[[0], [0]])

        return S, lambdas


    # Optimize F, keeping S fixed.
    def optimize_F(self, S):
        LS = normalized_laplacian(S)    
        return eigenvectors(LS, num=self.num_clusters)


    # Optimize S, keeping W and F fixed.
    def optimize_S(self, W, F):
        Q = distance_matrix(F, metric='euclidean')
        P = (2*W - self.gamma*Q)/(2 + 2*self.alpha)

        S = np.zeros((self.num_nodes, self.num_nodes)) 
        for index in range(S.shape[0]):
            S[index] = best_simplex_projection(P[index])
        
        return S
    
    # Optimize lambdas, keeping S fixed.
    def optimize_lambdas(self, S, init_lambdas):

        # The objective function, quadratic in lambda.
        def objective(lambdas):
            W = np.tensordot(lambdas, self.similarity_matrices, axes=[[0], [0]])
            value = np.trace(np.matmul(W.T, W))
            value -= 2 * np.trace(np.matmul(S.T, W))
            value += self.beta * np.dot(lambdas, lambdas)
            return value
    
        # Constraints are non-negativity and 1-summation.
        # Non-negativity is handled by the bounds() function. 
        def constraints():
            def sum_one(lambdas):
                return np.sum(lambdas) - 1

            return  {
                'type': 'eq',
                'fun': sum_one,
            }

        # Handle non-negativity
        def bounds(init_lambdas):
            return [(0, 1) for init_lambda in init_lambdas]

        return minimize(objective, init_lambdas, method='SLSQP', constraints=constraints(), bounds=bounds(init_lambdas)).x

